//############################################################################
// ~ International Laser Display Association ~
// ~     ILDA file format driver in C++      ~
// ~             built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_ilda.hpp"

//############################################################################
ez_ilda_segment::ez_ilda_segment(ez_ilda_point p0,
                                 ez_ilda_point p1,
                                 float         x_rotation,
                                 string        font,
                                 string        text
                                )
                 : palette_index(EZ_ILDA_DEFAULT),
                   error        (EZ_ILDA_OK     )
{   // 3D text constructor
    ez_ilda_frame_set font_frames;
    font_frames.from_file(font);
    if(font_frames.number_of_frames() < ('~' - '!' + 1))
        error = EZ_ILDA_INCOMPATIBLE_FORMATS;
    else
    {
        ez_real_segment real_segment;
        ez_3D_float     float_3D;
        float           length     = p0 | p1,
                           y_rotation = atan2(p1.z - p0.z, p1.x - p0.x);
        int             text_index ,
                        point_index;
        ez_ilda_point   p1_minus_y_rotation = ez_ilda_point(rotate_point_around_point_y(p1, p0, -y_rotation), (ez_color)p1);
        float z_rotation = atan2(p1_minus_y_rotation.y - p0.y, p1_minus_y_rotation.x - p0.x);

        while(x_rotation < 0)
            x_rotation += 360;
        x_rotation *= one_degree; // convert to radians!

        p1.unblank();
        for(text_index = 0; text_index < (int)text.size(); text_index++)
        {
            if(text[text_index] >= '!' && text[text_index] <= '~')
            {
                for(point_index = 0; point_index < (int)font_frames[text[text_index] - '!'].size(); point_index++)
                {
                    float_3D   = font_frames[text[text_index] - '!'].at(point_index);
                    float_3D.x = float_3D.x + (text_index * 24000.0);
                    real_segment.push_back(ez_real_point(float_3D, (ez_color)p1, p1.k, p1.c));                    
                    if(font_frames[text[text_index] - '!'].at(point_index).is_blank())
                        real_segment.back().blank();
                }
            }
        }
        real_segment.rotate_around_point(p0, ez_3D_float(x_rotation, y_rotation, z_rotation));
        for(point_index = 0; point_index < (int)real_segment.size(); point_index++)
        {
            push_back(ez_ilda_point(   real_segment[point_index]
                                     * (length / real_segment.segment_width())
                                     + p0
                                   )
                     );
            back().c = ez_ilda_pointer->selected_palette_color;
            back().r = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].r;
            back().g = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].g;
            back().b = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].b;
        }
    }
}

//############################################################################
ez_ilda_segment::ez_ilda_segment(ez_ilda_point center,
                                 ez_ilda_point arc_start,
                                 float arc_angle
                                )
                 : palette_index(EZ_ILDA_DEFAULT),
                   error        (EZ_ILDA_OK     )
{   // 3D circular arc constructor
    while(arc_angle <= 0) arc_angle += 360;

    int           total_points = (int)ceil(arc_angle / ez_ilda_pointer->rendered_arc_angle);
    float         arc_step     = arc_angle / total_points,
                  radius       = center | arc_start,
                     y_rotation   = atan2(arc_start.z - center.z, arc_start.x - center.x);

    ez_ilda_point point(arc_start);
    push_back(ez_ilda_point(   rotate_point_around_point_y(point, center, -y_rotation)
                             , (ez_color)point
                           )
             );
    point.unblank();
    float z_rotation = atan2(back().y - center.y, back().x - center.x);
    arc_step *= one_degree; // convert to radians!
    for(int a = 1; a <= total_points; a++)
    {
        point.x = (short)(radius * cos(a * arc_step + z_rotation));
        point.y = (short)(radius * sin(a * arc_step + z_rotation));
        point.z = center.z;        
        push_back(ez_ilda_point(point + center));        
        back().c = ez_ilda_pointer->selected_palette_color;
        back().r = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].r;
        back().g = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].g;
        back().b = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].b;
    }
    rotate_around_point_y(center, y_rotation);
}

//############################################################################
ez_ilda_segment::ez_ilda_segment(ez_ilda_point center,
                                 ez_ilda_point arc_start,
                                 float arc_angle,
                                 float radii_ratio
                                )
                : palette_index(EZ_ILDA_DEFAULT),
                  error        (EZ_ILDA_OK     )
{   // 3D elliptical arc constructor
    while(arc_angle <= 0)
        arc_angle += 360;

    int           total_points = (int)ceil(arc_angle / ez_ilda_pointer->rendered_arc_angle);
    float         arc_step     = arc_angle / total_points,
                  radius_1     = center | arc_start,
                  radius_2     = radius_1 * radii_ratio,
                     y_rotation   = atan2(arc_start.z - center.z, arc_start.x - center.x);
    ez_ilda_point point(ez_ilda_point(   rotate_point_around_point_y(arc_start, center, -y_rotation)
                                       , (ez_color)arc_start
                                     )
                       );
    float z_rotation = atan2(point.y - center.y, point.x - center.x);
    point.x = (short)(radius_1);
    point.y = 0;
    point.z = center.z;
    push_back(ez_ilda_point(point + center));
    point.unblank();
    arc_step *= one_degree; // convert to radians!
    for(int a = 1; a <= total_points; a++)
    {
        point.x = (short)(radius_1 * cos(a * arc_step));
        point.y = (short)(radius_2 * sin(a * arc_step));
        point.z = center.z;
        push_back(ez_ilda_point(point + center));
        back().c = ez_ilda_pointer->selected_palette_color;
        back().r = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].r;
        back().g = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].g;
        back().b = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].b;
    }
    rotate_around_point_z(center, z_rotation);
    rotate_around_point_y(center, y_rotation);
}


//############################################################################
ez_ilda_segment::ez_ilda_segment(ez_ilda_point center,
                                 ez_ilda_point first_vertex,
                                 int           number_of_sides
                                )
                : palette_index      (EZ_ILDA_DEFAULT),
                  error              (EZ_ILDA_OK     )
{   // 3D polygon constructor
    float   arc_step   = two_pi / number_of_sides,
            radius     = center | first_vertex,
            y_rotation = atan2(first_vertex.z - center.z, first_vertex.x - center.x);

    ez_ilda_point point(first_vertex);
    push_back(ez_ilda_point(   rotate_point_around_point_y(   point
                                                            , center
                                                            , -y_rotation
                                                          )
                             , (ez_color)point
                           )
             );
    point.unblank();
    float z_rotation = atan2(back().y - center.y, back().x - center.x);
    for(int a = 1; a <= number_of_sides; a++)
    {
        point.x = (short)(radius * cos(a * arc_step + z_rotation));
        point.y = (short)(radius * sin(a * arc_step + z_rotation));
        point.z = center.z;
        push_back(ez_ilda_point(point + center));
        back().c = ez_ilda_pointer->selected_palette_color;
        back().r = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].r;
        back().g = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].g;
        back().b = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].b;                            
    }
    rotate_around_point_y(center, y_rotation);
}

//############################################################################
ez_3D_short ez_ilda_segment::segment_front() const
{
    ez_3D_short front_;
    front_.z = EZ_MIN_SHORT;
    for(int i = 0; i < number_of_points(); i++)
        if(at(i).z > front_.z)
            front_ = at(i);
    return front_;
}

//############################################################################
ez_3D_short ez_ilda_segment::segment_back() const
{
    ez_3D_short back_;
    back_.z = EZ_MAX_SHORT;
    for(int i = 0; i < number_of_points(); i++)
        if(at(i).z < back_.z)
            back_ = at(i);
    return back_;
}

//############################################################################
ez_3D_short ez_ilda_segment::segment_top() const
{
    ez_3D_short top_;
    top_.y = EZ_MIN_SHORT;
    for(int i = 0; i < number_of_points(); i++)
        if(at(i).y > top_.y)
            top_ = at(i);
    return top_;
}

//############################################################################
ez_3D_short ez_ilda_segment::segment_bottom() const
{
    ez_3D_short bottom_;
    bottom_.y = EZ_MAX_SHORT;
    for(int i = 0; i < number_of_points(); i++)
        if(at(i).y < bottom_.y)
            bottom_ = at(i);
    return bottom_;
}

//############################################################################
ez_3D_short ez_ilda_segment::segment_right() const
{
    ez_3D_short right_;
    right_.x = EZ_MIN_SHORT;
    for(int i = 0; i < number_of_points(); i++)
        if(at(i).x > right_.x)
            right_ = at(i);
    return right_;
}

//############################################################################
ez_3D_short ez_ilda_segment::segment_left() const
{
    ez_3D_short left_;
    left_.x = EZ_MAX_SHORT;
    for(int i = 0; i < number_of_points(); i++)
        if(at(i).x < left_.x)
            left_ = at(i);
    return left_;
}

//############################################################################
short ez_ilda_segment::segment_height() const
{
    return (segment_top().y - segment_bottom().y);
}
//############################################################################
short ez_ilda_segment::segment_width() const
{
    return (segment_right().x - segment_left().x);
}

//############################################################################
short ez_ilda_segment::segment_depth() const
{
    return (segment_front().z - segment_back().z);
}

//############################################################################
short ez_ilda_segment::segment_size() const
{
    short greatest = 0;
    if(segment_width () > greatest)    greatest = segment_width ();
    if(segment_height() > greatest)    greatest = segment_height();
    if(segment_depth () > greatest)    greatest = segment_depth ();
    return greatest;
}

//############################################################################
bool ez_ilda_segment::is_closed_polygon() const
{
    if(   (first_lit_anchor() | last_lit_point())
        < ez_ilda_pointer->insignificant_distance
      )
        return true;
    return false;
}

//############################################################################
void ez_ilda_segment::blank_all_points()
{
    for(int i = 1; i < number_of_points(); i++)
        at(i).unblank();
    return;
}

//############################################################################
void ez_ilda_segment::unblank_all_points()
{
    for(int i = 1; i < number_of_points(); i++)
        at(i).unblank();
    return;
}

//############################################################################
//void ez_ilda_segment::remove_added_points()/
//{
//    if(number_of_points() > 1)
//    {
//        ez_ilda_segment reduced;
//        reduced.push_back(front());
//        for(int i = 1; i < number_of_points() - 1; i++)
//        {
//            if(at(i).is_unique())
//                reduced.push_back(at(i));
//        }
//        *this = reduced;
//    }
//    return;
//}

//############################################################################
ez_ilda_segment& ez_ilda_segment::reverse()
{
    int               i;
    ez_ilda_segment   reversed;
    ez_ilda_point     point;

    if(number_of_points() > 2)
    {
        i = number_of_points() - 1;
        //--------------------------------------------------------------------
        point = at(i--);
        point.blank();
        reversed.push_back(point);       
        //--------------------------------------------------------------------
        while(i >= 0)
        {
            point   = at(i    );
            point.r = at(i + 1).r;
            point.g = at(i + 1).g;
            point.b = at(i + 1).b;
            point.c = at(i + 1).c;
            point.k = at(i + 1).k;
            reversed.push_back(point);
            i--;
        }
        //--------------------------------------------------------------------
        *this = reversed;
    }
    else if(number_of_points() == 2)
    {
        //--------------------------------------------------------------------
        point = at(1);
        point.blank();
        reversed.push_back(point); 
        point = at(0);
        point.r = at(1).r;
        point.g = at(1).g;
        point.b = at(1).b;
        point.c = at(1).c;
        point.k = at(1).k;
        reversed.push_back(point); 
        *this = reversed;
    }
    return *this;
}

//############################################################################
ez_ilda_segment ez_ilda_segment::shortest_path_of_segments()
{
    ez_ilda_frame_set segments       = find_segments();
    int               total_segments = segments.number_of_frames();
    //------------------------------------------------------------------------
    if(total_segments > 1)
    {
        vector<bool>             been_here                              ;
        int                      i                                      ,
                                 segment_index_index                    ,
                                 points_away                            ,
                                 points_away_temp                       ,
                                 closest_segment_index = 0              ;
        ez_3D_float              closest_to_center(40000, 40000, 40000) ; // more positive than possible
        ez_ilda_segment          reordered(palette_index)               ;
        //--------------------------------------------------------------------
        for(i = 0; i < total_segments; i++)
        {
            been_here.push_back(false);

            if(    fabs(closest_to_center.sum())
                >= fabs(segments[i].first_lit_anchor().sum())
              )
            {
                closest_to_center = segments[i].first_lit_anchor();
                closest_segment_index = (i + 1); // can not be zero
            }
            if(    fabs(closest_to_center.sum())
                 > fabs(segments[i].last_lit_point().sum())
              )
            {
                closest_to_center = segments[i].last_lit_point();
                closest_segment_index = -(i + 1);
            }
        }
        //--------------------------------------------------------------------
        if(closest_segment_index > 0)
        {
            reordered = segments[closest_segment_index - 1];
            been_here[closest_segment_index - 1] = true;
        }
        else // (closest_segment_index < 0)
        {
            reordered = (segments[(-closest_segment_index) - 1].reverse());
            been_here[(-closest_segment_index) - 1] = true;
        }
        //--------------------------------------------------------------------
        for(segment_index_index = 1; segment_index_index < total_segments; segment_index_index++)
        {
            points_away           = LONG_MAX;
            closest_segment_index = 0;
            for(i = 0; i < total_segments; i++)
            {
                if(!been_here[i])
                {
                    points_away_temp =    distance_points(reordered.last_lit_point(),
                                                          segments[i].first_lit_anchor()
                                                         )
                                       +  dwell_points(delta_angle((reordered.last_lit_anchor() | segments[i].first_lit_anchor()),
                                                                   (reordered.last_lit_anchor() | reordered.last_lit_point ()),
                                                                   (reordered.last_lit_point () | segments[i].first_lit_anchor())
                                                                  )
                                                      )
                                       ;
                    //--------------------------------------------------------
                    if(points_away_temp < points_away)
                    {
                        points_away = points_away_temp;
                        closest_segment_index = (i + 1);
                    }
                    //--------------------------------------------------------
                    points_away_temp =    distance_points(reordered.last_lit_point(),
                                                          segments[i].last_lit_point()
                                                         )
                                       +  dwell_points(delta_angle((reordered.last_lit_anchor() | segments[i].last_lit_point()),
                                                                   (reordered.last_lit_anchor() | reordered.last_lit_point ()),
                                                                   (reordered.last_lit_point () | segments[i].last_lit_point())
                                                                  )
                                                      )
                                       ;
                    //--------------------------------------------------------
                    if(points_away_temp < points_away)
                    {
                        points_away = points_away_temp;
                        closest_segment_index = -(i + 1);
                    }
                    //--------------------------------------------------------
                }
            }
            if(closest_segment_index > 0)
            {
                reordered += segments[closest_segment_index - 1];
                been_here[closest_segment_index - 1] = true;
            }
            else // (closest_segment_index < 0)
            {
                reordered += (segments[(-closest_segment_index) - 1].reverse());
                been_here[(-closest_segment_index) - 1] = true;
            }
        }
        //--------------------------------------------------------------------
        return reordered;
        //--------------------------------------------------------------------
    }
    else
        return *this;
}

//############################################################################
ez_ilda_segment ez_ilda_segment::zoof_sort_segments()
{
    return *this;
}

//############################################################################
void ez_ilda_segment::minimize_blank_points()
{
    if(number_of_points() > 1)
    {
        int               i;
        ez_ilda_segment   minimum_blanking(palette_index);
        //--------------------------------------------------------------------
        minimum_blanking.push_back(front());
        for(i = 1; i < number_of_points(); i++)
        {
            if(    at(i).is_blank()
                && minimum_blanking.back().is_blank()
              )
                minimum_blanking.back() = at(i);
            else
                minimum_blanking.push_back(at(i));
        }
        //--------------------------------------------------------------------
        if(minimum_blanking.back().is_blank())
            minimum_blanking.pop_back();
        *this = minimum_blanking;
    }
    return;
}

//############################################################################
void ez_ilda_segment::remove_dwell_points()
{
    if(number_of_points() > 1)
    {
        int               i;
        ez_ilda_segment   no_black_points(palette_index);
        ez_ilda_segment   no_dwell_points(palette_index);
        //--------------------------------------------------------------------
        no_black_points.push_back(front());
        for(i = 1; i < number_of_points(); i++)
        {
            if(!at(i).is_black())
                no_black_points.push_back(at(i));
        }
        //--------------------------------------------------------------------
        no_dwell_points.push_back(no_black_points.front());
        for(i = 1; i < no_black_points.number_of_points(); i++)
        {
            if(    no_black_points[i]
                != no_dwell_points.back()
              )
                no_dwell_points.push_back(no_black_points[i]);
        }
        //--------------------------------------------------------------------
        *this = no_dwell_points;
    }
    return;
}

//############################################################################
void ez_ilda_segment::remove_equivelant_segments()
{
//    ez_ilda_frame_set segments = find_segments();
    return;
}

//############################################################################
void ez_ilda_segment::minimize_vectors()
{
    if(number_of_points() > 1)
    {
        int               i;
        ez_ilda_segment   minimum_vectors(palette_index);
        //--------------------------------------------------------------------
        minimum_vectors.push_back(front());
        //--------------------------------------------------------------------
        for(i = 1; i < number_of_points() - 1; i++)
        {
            if(    (    next_point_angle(i)
                     >  ez_ilda_pointer->zero_angle_tolerance
                   )
                || (    at(i    ).color()
                     != at(i + 1).color()
                   )
              )
                minimum_vectors.push_back(at(i));
        }
        minimum_vectors.push_back(at(i));
        *this = minimum_vectors;
    }
    return;
}

//############################################################################
void ez_ilda_segment::minimize_points()
{
    minimize_blank_points ();
    remove_dwell_points   ();
    bond_segments         ();
    minimize_vectors      ();
    return;
}

//############################################################################
void ez_ilda_segment::accentuate_the_positive()
{
    if(number_of_points() > 1)
    {
        bool  is_all_positive = true;
        int   i;
        //--------------------------------------------------------------------
        if(is_2D())
        {        
            for(i = 0; i < number_of_points(); i++)
                if(at(i).x < 0 || at(i).y < 0)
                    is_all_positive = false;
            if(is_all_positive)
                for(i = 0; i < number_of_points(); i++)
                {
                    at(i).x = (at(i).x - EZ_HALF_SHORT) * 2;
                    at(i).y = (at(i).y - EZ_HALF_SHORT) * 2;
                }
        }
        else
        {                 
            for(i = 0; i < number_of_points(); i++)
                if(at(i).x < 0 || at(i).y < 0 || at(i).z < 0)
                    is_all_positive = false;                
            if(is_all_positive)
                for(i = 0; i < number_of_points(); i++)
                {
                    at(i).x = (at(i).x - EZ_HALF_SHORT) * 2;
                    at(i).y = (at(i).y - EZ_HALF_SHORT) * 2;
                    at(i).z = (at(i).z - EZ_HALF_SHORT) * 2;
                }
        }        
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ez_ilda_segment::eliminate_the_negative()
{
    if(number_of_points() > 1)
    {
        bool              is_all_positive = true;
        int               i;
        //--------------------------------------------------------------------
        if(is_2D())
        {        
            for(i = 0; i < number_of_points(); i++)
            {
                at(i).x = at(i).x / 2 + EZ_HALF_SHORT;
                at(i).y = at(i).y / 2 + EZ_HALF_SHORT;
            }
        }
        else
        {                 
            for(i = 0; i < number_of_points(); i++)
            {
                at(i).x = at(i).x / 2 + EZ_HALF_SHORT;
                at(i).y = at(i).y / 2 + EZ_HALF_SHORT;
                at(i).z = at(i).z / 2 + EZ_HALF_SHORT;
            }
        }        
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
ez_ilda_segment& ez_ilda_segment::reorder_segments()
{
    *this = shortest_path_of_segments();
    return *this;
}

//############################################################################
ez_ilda_segment& ez_ilda_segment::fracture_segments()
{
    if(number_of_points() > 1)
    {
        int               point_index;
        ez_ilda_segment   fractured(palette_index);
        //--------------------------------------------------------------------
        for(point_index = 0; point_index < number_of_points(); point_index++)
        {
            fractured.push_back(at(point_index));
            fractured.push_back(at(point_index));
            fractured.back().blank();
        }
        //--------------------------------------------------------------------
        fractured.pop_back();
        *this = fractured;
    }
    return *this;
}

//############################################################################
ez_ilda_segment& ez_ilda_segment::bond_segments()
{
    if(number_of_points() > 1)
    {
        int               point_index;
        ez_ilda_segment   bonded(palette_index);
        //--------------------------------------------------------------------
        reorder_segments();
        minimize_blank_points();
        bonded.push_back(front());
        //--------------------------------------------------------------------
        for(point_index = 1; point_index < number_of_points() - 1; point_index++)
        {
            if(     at(point_index).is_blank()
                && (at(point_index) | at(point_index - 1)) < ez_ilda_pointer->insignificant_distance
              )
            {;} // do nothing
            else
                bonded += at(point_index);
        }
        //--------------------------------------------------------------------
        bonded.push_back(back());
        *this = bonded;
    }
    return *this;
}

//############################################################################
ez_ilda_segment& ez_ilda_segment::add_new_point(int point_index)
{
    if(!size())
    {
        push_back(ez_ilda_point());
        push_back(ez_ilda_point());
    }
    else if(    point_index >= 0
             && point_index <  (int)size()
             && size()
           )
    {
        int              i;
        ez_ilda_segment  segment(palette_index);
        if(point_index < (int)size() - 1) // not the last point
        {
            for(i = 0; i <= point_index; i++)
                segment += at(i);

            if((at(point_index) | at(point_index + 1)) > ez_ilda_pointer->insignificant_distance)
                segment += ez_ilda_point(   scale_point_around_point(   at(point_index + 1)
                                                                      , at(point_index)
                                                                      , ez_3D_float(0.5, 0.5, 0.5)
                                                                    )
                                          , at(point_index) // is an rgb
                                          , at(point_index).k
                                          , at(point_index).c
                                        );

            for(i = point_index + 1; i < (int)size(); i++)
                segment += at(i);
        }
        else // it is the last point
        {
            segment = *this;
            segment += ez_ilda_point(   scale_point_around_point(   back()
                                                                  , at(size() - 2)
                                                                  , ez_3D_float(2.0, 2.0, 2.0)
                                                                )
                                      , back() // is an rgb
                                      , back().k
                                      , back().c
                                    );
        }
        *this = segment;
    }
    return *this;
}

//############################################################################
ez_ilda_segment& ez_ilda_segment::break_segment(int& point_index)
{
    if(    point_index
        && point_index < (int)size() - 1
      )
    {
        ez_ilda_segment  segment(palette_index);
        int              i;

        for(i = 0; i <= point_index; i++)
            segment += at(i);

        segment += at(point_index);
        segment.back().blank();

        for(i = point_index + 1; i < (int)size(); i++)
            segment += at(i);

        point_index += 2;
        *this = segment;
    }
    return *this;
}

//############################################################################
ez_ilda_segment& ez_ilda_segment::connect_the_dots(int p0, int p1)
{
    if(    p0 != p1
        && p0 > 0
        && p0 < (int)size()
        && p1 > 0
        && p1 < (int)size()
        && (  (   at(p0)
                | at(p1)
              )
            > ez_ilda_pointer->insignificant_distance
           )
      )
    {                                
        push_back(at(p0));
        back().blank();        
        push_back(at(p1));
        back().unblank();
        back().c = ez_ilda_pointer->selected_palette_color;
        back().r = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].r;
        back().g = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].g;
        back().b = ez_ilda_pointer->palette_picker(ez_ilda_pointer->palette_index)[ez_ilda_pointer->selected_palette_color].b;
    }    
    return *this;
}

//############################################################################
double ez_ilda_segment::next_point_angle(int point_index) const
{
    double angle = 0.0;
    if(point_index != 0)
    {
        if(number_of_points() > 1)
        {
            int _0  = 0,
                _1  = 0,
                _2  = 0;
           //--------------------------------------------------------------------
            if(point_index > 0 && point_index < number_of_points() - 1)
            {
                _0 = point_index - 1;
                _1 = point_index    ;
                _2 = point_index + 1;
            }
           //--------------------------------------------------------------------
            else if(point_index == number_of_points() - 1)
            {
                _0 = number_of_points() - 2;
                _1 = 0                     ;
                _2 = 1                     ;
            }
            //--------------------------------------------------------------------
            angle = delta_angle( at(_0) | at(_2), // distance between points 0,2
                                 at(_0) | at(_1),
                                 at(_1) | at(_2)
                               );
        }
    }
    return angle;
}

//############################################################################
double ez_ilda_segment::total_angle() const
{
    double total_angle = 0.0;
    int    i;
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_points() - 1; i++)
        total_angle += next_point_angle(i);
    return total_angle;
}

//############################################################################
double ez_ilda_segment::max_angle() const
{
    double max_angle = 0.0;
    int    i;
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_points() - 1; i++)
        if(max_angle < next_point_angle(i))
            max_angle = next_point_angle(i);
    return max_angle;
}

//############################################################################
double ez_ilda_segment::next_point_distance(int point) const
{
    if(number_of_points() > 1)
    {
        int _1 = 0,
            _2 = 0;
        //--------------------------------------------------------------------
        if(point >= 0 && point < number_of_points() - 1)
        {
            _1 = point    ;
            _2 = point + 1;
        }
        else if(point == number_of_points() - 1)
        {
            _1 = number_of_points() - 1;
            _2 =          0;
        }
        //--------------------------------------------------------------------
        return at(_1) | at(_2);
    }
    return 0;
}

//############################################################################
double ez_ilda_segment::total_distance() const
{
    double total_distance = 0.0;
    int    i;
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_points() - 1; i++)
        total_distance += next_point_distance(i);
    return total_distance;
}

//############################################################################
double ez_ilda_segment::max_distance() const
{
    double max_distance = 0.0;
    int    i;
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_points() - 1; i++)
        if(max_distance < next_point_distance(i))
            max_distance = next_point_distance(i);
    return max_distance;
}

//############################################################################
void ez_ilda_segment::add_dwell_points()
{
    if(number_of_points() > 1)
    {
        ez_ilda_point    point,
                         black_point;
        ez_ilda_segment  angle_optimized(palette_index);

        int              i                 ,
                         dwell_point_count ,
                         dwell_point_total ;
        //--------------------------------------------------------------------
        angle_optimized.push_back(front());
        for(i = 1; i < number_of_points(); i++)
        {
            angle_optimized.push_back(at(i));
            dwell_point_total = dwell_points(next_point_angle(i));
            if(dwell_point_total)
            {
                point = black_point = at(i);                
                if(black_point.color() && ez_ilda_pointer->black_dwell_points)
                {
                    black_point.c = ez_ilda_pointer->palette_picker(palette_index).black;
                    black_point.r =
                    black_point.g =
                    black_point.b = 0x00;
                }
                point.clear_unique();
                black_point.clear_unique();
                for(dwell_point_count = 0; dwell_point_count < dwell_point_total; dwell_point_count++)
                    if(dwell_point_count < ez_ilda_pointer->lit_dwell_overhang)
                        angle_optimized.push_back(point);
                    else
                        angle_optimized.push_back(black_point);
            }
        }
        //--------------------------------------------------------------------
        *this = angle_optimized;
    }
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_ilda_segment::add_span_points()
{
    if(number_of_points() > 1)
    {
        int              i;
        ez_ilda_segment  distance_optimized(palette_index);
        //--------------------------------------------------------------------
        for(i = 0; i < number_of_points() - 1; i++)
        {
            distance_optimized.push_back(at(i));
            if(next_point_distance(i) > ez_ilda_pointer->distance_delta_max)
                distance_optimized += ez_ilda_segment(at(i), at(i + 1));
        }
        //--------------------------------------------------------------------
        *this = distance_optimized;
    }
    return;
}

//############################################################################
void ez_ilda_segment::flip(int plane)
{
    if(number_of_points() > 1)
    {
        int i;
        switch(plane)
        {
           case 0: // X
                   for(i = 0; i < number_of_points(); i++)
                       at(i).x = -at(i).x;
                   break;                                
           //-----------------------------------------------------------------
           case 1: // Y
                   for(i = 0; i < number_of_points(); i++)
                       at(i).y = -at(i).y;
                   break;
           //-----------------------------------------------------------------
           case 2: // Z
                   for(i = 0; i < number_of_points(); i++)
                       at(i).z = -at(i).z;
                   break;
           //-----------------------------------------------------------------
           case 3: // X, Y
                   for(i = 0; i < number_of_points(); i++)
                   {
                       at(i).x = -at(i).x;
                       at(i).y = -at(i).y;
                   }
                   break;                                
           //-----------------------------------------------------------------
           case 4: // X, Y, Z
                   for(i = 0; i < number_of_points(); i++)
                       at(i) = -at(i);
                   break;                                
           //-----------------------------------------------------------------
        }
    }
    return;
}

//############################################################################
void ez_ilda_segment::quarter_turn(int plane, int turns)
{
    if(number_of_points() > 1)
    {
        int   i, j;
        short temp;
        switch(plane)
        {
           case 0: // X Y
                   for(i = 0; i < number_of_points(); i++)
                       for(j = 0; j < turns; j++)
                       {
                           temp = -at(i).x;
                           at(i).x = at(i).y;
                           at(i).y = temp;
                       }
                   break;                                
           //-----------------------------------------------------------------
           case 1: // Z Y
                   for(i = 0; i < number_of_points(); i++)
                       for(j = 0; j < turns; j++)
                       {
                           temp = -at(i).z;
                           at(i).z = at(i).y;
                           at(i).y = temp;
                       }
                   break;
           //-----------------------------------------------------------------
           case 2: // X Z
                   for(i = 0; i < number_of_points(); i++)
                       for(j = 0; j < turns; j++)
                       {
                           temp = -at(i).x;
                           at(i).x = at(i).z;
                           at(i).z = temp;
                       }
                   break;
           //-----------------------------------------------------------------
        }
    }
    return;
}

//############################################################################
void ez_ilda_segment::z_order_points(unsigned short span)
{
    if(number_of_points() > 1)
    {
        int i,
            step = span / number_of_points();
        for(i = 0; i < number_of_points(); i++)
            at(i).z = (short)((i * step) - (span / 2));
    }
    return;
}

//############################################################################
void ez_ilda_segment::flatten_z()
{
    if(number_of_points() > 1)
    {
        for(int i = 0; i < number_of_points(); i++)
            at(i).z = 0;
    }
    return;
}

//############################################################################
int ez_ilda_segment::rotate(ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        int            i,
                       out_of_bounds;
        ez_3D_float    center = find_center();

        for(i = 0; i < number_of_points(); i++)
        {
            out_of_bounds = ez_ilda_bounds_check(rotate_point_around_point(at(i), center, a), EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        for(i = 0; i < number_of_points(); i++)
            at(i) = ez_ilda_point( rotate_point_around_point(at(i), center, a),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::rotate_around_origin(ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        int            i,
                       out_of_bounds;

        //ez_3D_short    center;

        for(i = 0; i < number_of_points(); i++)
        {
            out_of_bounds = ez_ilda_bounds_check(rotate_point(at(i), a), EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        for(i = 0; i < number_of_points(); i++)
            at(i) = ez_ilda_point( rotate_point(at(i), a),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::rotate_around_point(ez_3D_float p, ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        int            i,
                       out_of_bounds;

        for(i = 0; i < number_of_points(); i++)
        {
            out_of_bounds = ez_ilda_bounds_check(rotate_point_around_point(at(i), p, a), EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        for(i = 0; i < number_of_points(); i++)
            at(i) = ez_ilda_point( rotate_point_around_point(at(i), p, a),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::rotate_around_point_x(ez_3D_float p, float a)
{
    if(number_of_points() > 1)
    {
        int            i,
                       out_of_bounds;

        for(i = 0; i < number_of_points(); i++)
        {
            out_of_bounds = ez_ilda_bounds_check(rotate_point_around_point_x(at(i), p, a), EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        for(i = 0; i < number_of_points(); i++)
            at(i) = ez_ilda_point( rotate_point_around_point_x(at(i), p, a),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::rotate_around_point_y(ez_3D_float p, float a)
{
    if(number_of_points() > 1)
    {
        int            i,
                       out_of_bounds;

        for(i = 0; i < number_of_points(); i++)
        {
            out_of_bounds = ez_ilda_bounds_check(rotate_point_around_point_y(at(i), p, a), EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        for(i = 0; i < number_of_points(); i++)
            at(i) = ez_ilda_point( rotate_point_around_point_y(at(i), p, a),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::rotate_around_point_z(ez_3D_float p, float a)
{
    if(number_of_points() > 1)
    {
        int            i,
                       out_of_bounds;

        for(i = 0; i < number_of_points(); i++)
        {
            out_of_bounds = ez_ilda_bounds_check(rotate_point_around_point_z(at(i), p, a), EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        for(i = 0; i < number_of_points(); i++)
            at(i) = ez_ilda_point( rotate_point_around_point_z(at(i), p, a),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::rotate_around_origin()
{
    return rotate_around_origin(ez_ilda_pointer->view_angle);
}

//############################################################################
void ez_ilda_segment::ripple(int direction, float amplitude, float freq, float phase)
{
    if(number_of_points() > 1)
    {
        int i;
        switch(direction)
        {
            case 0: for(i = 0; i < number_of_points(); i++) // parellel through x
                        at(i).z = short(amplitude * (16384.0 / pi) * sin(at(i).x / 32768.0 * two_pi * freq + phase));
                    break;
            //----------------------------------------------------------------
            case 1: for(i = 0; i < number_of_points(); i++) // parellel through y
                        at(i).z = short(amplitude * (16384.0 / pi) * sin(at(i).y / 32768.0 * two_pi * freq + phase));
                    break;
            //----------------------------------------------------------------
            case 2: for(i = 0; i < number_of_points(); i++) // x * y ~ egg carton effect
                        at(i).z = short(amplitude * (16384.0 / pi) * sin(at(i).x * at(i).y / 1073741824.0 * two_pi * freq + phase));
                    break;
            //----------------------------------------------------------------
            case 3: for(i = 0; i < number_of_points(); i++) // circular, concentric to the origin
                        at(i).z = short(amplitude * (16384.0 / pi) * sin(sqrt(float(at(i).x * at(i).x + at(i).y * at(i).y)) / 32768.0 * two_pi * freq + phase));
            //----------------------------------------------------------------
        }
    }
    return;
}

//############################################################################
void ez_ilda_segment::strip_color()
{
    palette_index = EZ_ILDA_DEFAULT;
    if(size() > 1)
        for(int i = 0; i < (int)size(); i++)
        {
            if(at(i).k & EZ_ILDA_BLANKING_BIT)
            {
                at(i).c = 0;
                at(i).r = 0;
                at(i).g = 0;
                at(i).b = 0;
            }
            else
            {
                at(i).c =  55; // EZ_ILDA_DEFAULT white
                at(i).r = 255;
                at(i).g = 255;
                at(i).b = 255;
            }
        }
    ez_ilda_pointer->palette_index = EZ_ILDA_DEFAULT;
    return;
};

//############################################################################
void ez_ilda_segment::accept_palette(int index)
{
    if(    index != EZ_TRUE_COLOR
        && index != palette_index
        && index <  ez_ilda_pointer->number_of_palettes()
        && index >= 0
      )
    {
        palette_index = index;
        set_rgb_from_palette();
        ez_ilda_pointer->palette_index = palette_index;
    }    
    return;
};

//############################################################################
void ez_ilda_segment::accept_target_palette()
{
    if(    ez_ilda_pointer->target_palette_index != EZ_TRUE_COLOR
        && ez_ilda_pointer->target_palette_index != palette_index
        && ez_ilda_pointer->target_palette_index <  ez_ilda_pointer->number_of_palettes()
        && ez_ilda_pointer->target_palette_index >= 0
      )
    {
        palette_index = ez_ilda_pointer->target_palette_index;
        set_rgb_from_palette();
        ez_ilda_pointer->palette_index = palette_index;
    }    
    return;
};


//############################################################################
void ez_ilda_segment::best_match_palette(int index)
{
    if(    index != palette_index
        && index != EZ_TRUE_COLOR
        && index <  ez_ilda_pointer->number_of_palettes()
        && index >= 0
      )
    {
        for(int i = 0; i < (int)size(); i++)
             at(i).c = ez_ilda_pointer->
                       palette_picker(index)
                       .best_match(at(i));
        palette_index = index;
        set_rgb_from_palette();
        ez_ilda_pointer->palette_index = palette_index;
    }    
    return;
};

//############################################################################
void ez_ilda_segment::best_match_target_palette()
{
    if(    ez_ilda_pointer->target_palette_index != palette_index
        && ez_ilda_pointer->target_palette_index <  ez_ilda_pointer->number_of_palettes()
        && ez_ilda_pointer->target_palette_index >= 0
      )
    {
        for(int i = 0; i < (int)size(); i++)
             at(i).c = ez_ilda_pointer->
                       palette_picker(ez_ilda_pointer->target_palette_index)
                       .best_match(at(i));
        palette_index = ez_ilda_pointer->target_palette_index;
        set_rgb_from_palette();
        ez_ilda_pointer->palette_index = palette_index;
    }    
    return;
};

//############################################################################
void ez_ilda_segment::rainbow_recolor(int effect, float factor)
{
    if(    palette_index != EZ_TRUE_COLOR
        && number_of_points() > 1
      )
    {
        int    i,
               j,
               span   =    ez_ilda_pointer->palette_picker(palette_index).last
                         - ez_ilda_pointer->palette_picker(palette_index).first,
               offset =    ez_ilda_pointer->palette_picker(palette_index).first;
        u_char color;

        if(effect > 8) effect = 0;
        if(effect < 0) effect = 8;

        switch(effect)
        {
            case 0: for(i = 1; i < number_of_points(); i++) // span through x
                        at(i).c = (u_char)(at(i).x * factor / 65536.0 * span) % span + offset;
                    break;
           //-----------------------------------------------------------------
            case 1: for(i = 1; i < number_of_points(); i++) // span through y
                        at(i).c = (u_char)(at(i).y * factor / 65536.0 * span) % span + offset;
                    break;
           //-----------------------------------------------------------------
            case 2: for(i = 1; i < number_of_points(); i++) // span through z
                        at(i).c = (u_char)(at(i).z * factor / 65536.0 * span) % span + offset;
                    break;
           //-----------------------------------------------------------------
            case 3: for(i = 1; i < number_of_points(); i++) // span through points
                        at(i).c = (u_char)((i * factor / (float)number_of_points()) * span) % span + offset;
                    break;
           //-----------------------------------------------------------------
            case 4: for(i = 1; i < number_of_points(); i++) // index through points
                        at(i).c = (u_char)(i % span + offset);
                    break;
           //-----------------------------------------------------------------
            case 5: // index through segments
                    {
                        ez_ilda_frame_set segments = find_segments();
                        for(i = 0; i < number_of_segments(); i++)
                        {
                            color = u_char(((i * factor) / number_of_segments()) * span) % span + offset;
                            for(j = 0; j < segments[i].number_of_points() - 1; j++)
                                segments[i].at(j).c = color;
                        }
                        *this = segments.sum_of_frames();
                    }
                    break;
           //-----------------------------------------------------------------
            case 6: // span through segments
                    {
                        ez_ilda_frame_set segments = find_segments();
                        for(i = 0; i < number_of_segments(); i++)
                        {
                            color = (i % span) + offset;
                            for(j = 0; j < segments[i].number_of_points() - 1; j++)
                                segments[i].at(j).c = color;
                        }
                        *this = segments.sum_of_frames();
                    }
                    break;
           //-----------------------------------------------------------------
            case 7: for(i = 1; i < number_of_points() - 1; i++)
                        at(i).c = (rand() % span) + offset;
                    break;
           //-----------------------------------------------------------------
            case 8: // random color segments
                    {
                        ez_ilda_frame_set segments = find_segments();
                        for(i = 0; i < number_of_segments(); i++)
                        {
                            color = (rand() % span) + offset;
                            for(j = 0; j < segments[i].number_of_points() - 1; j++)
                                segments[i].at(j).c = color;
                        }
                        *this = segments.sum_of_frames();
                    }
                    break;
           //-----------------------------------------------------------------
        }
        set_rgb_from_palette();
    }
    return;
}

//############################################################################
void ez_ilda_segment::rotate_colors()
{
    if(    palette_index != EZ_TRUE_COLOR
        && number_of_points() > 1
      )
    {
        int offset =   ez_ilda_pointer->palette_picker(palette_index).first,

            span   =   ez_ilda_pointer->palette_picker(palette_index).last
                     - ez_ilda_pointer->palette_picker(palette_index).first;
        //--------------------------------------------------------------------
        for(int i = 0; i < number_of_points(); i++)
            at(i).c = ((at(i).c - offset + 1) % span) + offset;
        set_rgb_from_palette();
    }
    return;
}

//############################################################################
bool ez_ilda_segment::find_rgb_in_palette(const ez_palette& palette)
{
    int           i,
                  j;
    vector<bool>  match(number_of_points(), false);
    bool          all_colors_found = true;
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_points(); i++)
        for(j = 0; j < palette.number_of_colors(); j++)
            if((ez_color)at(i) == palette.at(j))
            {
                at(i).c = (u_char)j;
                match[i] = true;
            }
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_points(); i++)
        all_colors_found &= match[i];
    //------------------------------------------------------------------------
    return all_colors_found;
}

//############################################################################
void ez_ilda_segment::set_rgb_from_palette()
{
    at(0).r = 
    at(0).g = 
    at(0).b = 0x00;
    //------------------------------------------------------------------------
    for(int i = 1; i < number_of_points(); i++)
    {
        at(i).r = ez_ilda_pointer->palette_picker(palette_index)[at(i).c].r;
        at(i).g = ez_ilda_pointer->palette_picker(palette_index)[at(i).c].g;
        at(i).b = ez_ilda_pointer->palette_picker(palette_index)[at(i).c].b;
    }
    return;
}

//############################################################################
void ez_ilda_segment::set_palette_to_332()
{
    at(0).c = 0x00;
    for(int i = 1; i < number_of_points(); i++)
        at(i).c =    (at(i).r & 0xe0)
                  | ((at(i).g & 0xe0) >> 3)
                  | ((at(i).b & 0xc0) >> 6);
    // does NOT set palette_index to EZ_EZFB_332
    return;
}

//############################################################################
void ez_ilda_segment::sync_rgb_and_palette()
{
    if(number_of_points() > 1)
    {
        if(palette_index != EZ_TRUE_COLOR)
            set_rgb_from_palette();
        //--------------------------------------------------------------------
        else
        {
            int            i                  ,
                           j                  ;
            char           palette_name[8]    ;
            ez_palette     palette            ;
            static int     palette_number = 0 ;
            //----------------------------------------------------------------
            sprintf(palette_name, "%07d", palette_number);
            palette.name = string("P") + palette_name;
            palette.push_back((ez_color)at(1));
            //----------------------------------------------------------------
            for(i = 1; i < number_of_points(); i++)
            {
                for(j = 0; j < palette.number_of_colors(); j++)
                    if(palette[j] == (ez_color)at(i))
                        break;
                if(j == palette.number_of_colors())
                    palette.push_back((ez_color)at(i));

                if(palette.number_of_colors() > EZ_PALETTE_MAX)
                    break;
            }
            //----------------------------------------------------------------
            if(palette.number_of_colors() <= EZ_PALETTE_MAX)
            {
                palette.reorder();
                palette.find_factors();
                find_rgb_in_palette(palette);
                ez_ilda_pointer->push_back_palette(palette);
                palette_index = ez_ilda_pointer->number_of_palettes() - 1;
                palette_number++;
            }
            //----------------------------------------------------------------
            else // (palette.number_of_colors() > EZ_PALETTE_MAX)
                set_palette_to_332();
        }
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ez_ilda_segment::bit_reduce_to_palette()
{
    if(palette_index == EZ_TRUE_COLOR)
    {
        set_palette_to_332();
        palette_index = EZ_EZFB_332;
        set_rgb_from_palette();
    }
    return;
}

//############################################################################
void ez_ilda_segment::promote_to_true_color()
{
    if(palette_index != EZ_TRUE_COLOR)
    {
        set_rgb_from_palette();
        palette_index = EZ_TRUE_COLOR;
    }
    return;
}

//############################################################################
void ez_ilda_segment::color_over_ezfb_bitmap(struct ezfb_bitmap* bmp)
{
    if(number_of_points() > 1)
    {
        int     i;
        u_int   color,
                factor = (65536 / bmp->xres);              
        //--------------------------------------------------------------------
        if(bmp->bpp <= 8)
            for(i = 1; i < number_of_points(); i++)
            {
                color = bmp->get_pixel(  bmp,
                                         (at(i).x + 32767) / factor,
                                         (at(i).y + 32767) / factor
                                      );
                at(i).r = bmp_get_palette_index_r(bmp, color);
                at(i).g = bmp_get_palette_index_g(bmp, color);
                at(i).b = bmp_get_palette_index_b(bmp, color);
            }
        //--------------------------------------------------------------------
        else
            for(i = 1; i < number_of_points(); i++)
            {
                color = bmp->get_pixel(  bmp,
                                         (at(i).x + 32767) / factor,
                                         (at(i).y + 32767) / factor
                                      );
                at(i).r = bmp->r_from_rgb(color);
                at(i).g = bmp->g_from_rgb(color);
                at(i).b = bmp->b_from_rgb(color);
            }
        //--------------------------------------------------------------------
        palette_index = EZ_TRUE_COLOR;
        sync_rgb_and_palette();
    }
    return;
}

//############################################################################
int ez_ilda_segment::move(ez_3D_float d)
{
    if(number_of_points() > 1)
    {
        ez_ilda_segment segment(*this);
        ez_3D_float     f;
        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            f = at(i).position();
            out_of_bounds = ez_ilda_bounds_check(f + d, EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            f = at(i).position();
            at(i) = ez_ilda_point( f + d,
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::move()
{
    return move(ez_ilda_pointer->view_offset);
}

//############################################################################
int ez_ilda_segment::scale(ez_3D_float s)
{
    ez_3D_float center = find_center();
    return scale_around_point(center, s);
}

//############################################################################
int ez_ilda_segment::scale_around_point(ez_3D_float p, ez_3D_float s)
{
    if(number_of_points() > 1)
    {
        ez_ilda_segment segment(*this);
        ez_3D_float     f;
        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            f = at(i).position();
            out_of_bounds = ez_ilda_bounds_check(((f - p) * s) + p, EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            f = at(i).position();
            at(i) = ez_ilda_point( ((f - p) * s) + p,
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::scale_around_origin(ez_3D_float f)
{
    if(number_of_points() > 1)
    {
        ez_ilda_segment segment(*this);
        int           i,
                      segment_index,
                      out_of_bounds = EZ_IN_BOUNDS;

        for(segment_index = 0; segment_index < segment.number_of_segments(); segment_index++)
            if((out_of_bounds = segment.scale_segment_around_origin(segment_index, f)))
                return out_of_bounds;

        for(i = 0; i < number_of_points(); i++)
            at(i) *= f;
        //--------------------------------------------------------------------
        //back() = front();
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::scale_around_origin()
{
    return scale_around_origin(ez_ilda_pointer->view_scale);
}

//############################################################################
ez_3D_float ez_ilda_segment::find_center() const
{
    ez_3D_float  center;
    if(number_of_points() > 1)
    {
        int            i             ,
                       max_x = -40000, // more negative than any short
                       min_x =  40000, // more positive than any short
                       max_y = -40000,
                       min_y =  40000,
                       max_z = -40000,
                       min_z =  40000;

        for(i = 0; i < number_of_points(); i++)
        {
            if(at(i).x > max_x)    max_x = at(i).x;
            if(at(i).x < min_x)    min_x = at(i).x;
            if(at(i).y > max_y)    max_y = at(i).y;
            if(at(i).y < min_y)    min_y = at(i).y;
            if(at(i).z > max_z)    max_z = at(i).z;
            if(at(i).z < min_z)    min_z = at(i).z;
        }
        center.x = max_x - ((max_x - min_x) / 2);
        center.y = max_y - ((max_y - min_y) / 2);
        center.z = max_z - ((max_z - min_z) / 2);
    }
    return center;
}

//############################################################################
int ez_ilda_segment::number_of_segments() const // a segment is a series of lit points
{
    int i,
        segment_count = 0;
    if(number_of_points() > 1)
    {
        for(i = 1; i < number_of_points(); i++)
        {
            if(at(i).is_lit())
            {
                while(at(i).is_lit() && i < (number_of_points() - 1))
                    i++;
                segment_count++;
            }
        }
    }
    return segment_count;
}

//############################################################################
bool ez_ilda_segment::find_segment(int segment_index, int& start, int& end) const
{   // the first segment is number zero!
    if(number_of_points() > 1)
    {
        int  i,
             segment_count = -1;
        //--------------------------------------------------------------------
        for(i = 1; i < number_of_points(); i++)
        {
            if(at(i).is_lit())
            {
                start = i - 1;
                while(at(i).is_lit() && i < (number_of_points() - 1))
                    i++;
                //------------------------------------------------------------
                end = i - 1;
                if(    i == (number_of_points() - 1)
                    && at(i).is_lit()
                  )
                {
                    end = i;
                }
                //------------------------------------------------------------
                segment_count++;
                if(segment_count == segment_index)
                    return true;
            }
        } // segment index either negative or out of range
        //--------------------------------------------------------------------
        if(segment_index > -1)
        {
            for(i = 1; i < number_of_points(); i++)
            {
                if(at(i).is_lit())
                {
                    start = i - 1;
                    while(at(i).is_lit() && i < (number_of_points() - 1))
                        i++;
                    end = i - 1;
                    if(    i == (number_of_points() - 1)
                        && at(i).is_lit()
                      )
                        end = i;
                    return false; // and set start, end to first segment
                }
            }
        }
    }
    //------------------------------------------------------------------------
    return false;
}

//############################################################################
bool ez_ilda_segment::find_segment(int point_index, int& start, int& end, int& segment_index) const
{
    start = end = segment_index = 0;
    if(number_of_points() > 1)
    {
        int  i,
             segment_count = 0; // the first segment is number zero!
        //--------------------------------------------------------------------
        for(i = 1; i < number_of_points() - 1; i++)
        {
            if(at(i).is_lit())
            {
                start = i - 1;
                while(at(i).is_lit() && i < (number_of_points() - 1))
                    i++;
                end = i - 1;

                if(    point_index >= start
                    && point_index <= end
                  )
                {
                    segment_index = segment_count;
                    return true;
                }

                if(point_index < start)
                {
                    segment_index = segment_count;
                    return false;
                }
                segment_count++;
            }
        }
    }
    //------------------------------------------------------------------------
    return false;
}

//############################################################################
ez_ilda_segment ez_ilda_segment::find_segment(int segment_index) const
{
    if(number_of_points() > 1)
    {
        int             start  ,
                        end    ;
        ez_ilda_segment segment(palette_index);
        find_segment(segment_index, start, end);
        for(int i = start; i <= end; i++)
            segment.push_back(at(i));
        return segment;
    }
    return NULL_segment();
}

//############################################################################
ez_ilda_frame_set ez_ilda_segment::find_segments() const
{
    ez_ilda_frame_set segments;
    //------------------------------------------------------------------------
    if(number_of_points() > 1)
    {
        int               i     ,
                          j     ,
                          start ,
                          end   ;
        ez_ilda_frame     frame(palette_index);
        //--------------------------------------------------------------------
        for(i = 1; i < number_of_points(); i++)
        {
            if(at(i).is_lit())
            {
                start = i - 1; // anchor to the first lit point
                while(at(i).is_lit() && i < (number_of_points() - 1))
                    i++;
                //------------------------------------------------------------
                end = i - 1;
                if(    i == (number_of_points() - 1)
                    && at(i).is_lit()
                  )
                    end = i;
                //------------------------------------------------------------
                frame.clear();
                for(j = start; j <= end; j++)
                    frame += at(j);
                segments.push_back(frame);
            }
        }
    }
    //------------------------------------------------------------------------
    return segments;
}

//############################################################################
ez_3D_float ez_ilda_segment::find_segment_center(int segment_index) const
{
    return (find_segment(segment_index)).find_center();
}

//############################################################################
int ez_ilda_segment::segment_index_of_point(int index) const
{
    int start, end, segment_index;
    find_segment(index, start, end, segment_index);
    return segment_index;
}

//############################################################################
int ez_ilda_segment::move_segment(int segment_index, ez_3D_float f)
{
    if(number_of_points() > 1)
    {
        int  i             ,
             out_of_bounds ,
             start         ,
             end           ;

        find_segment(segment_index, start, end);

        for(i = start; i <= end; i++)
        {
            out_of_bounds = ez_ilda_bounds_check(f + at(i), EZ_ILDA_CUBE);
            if(out_of_bounds != EZ_IN_BOUNDS)
                return out_of_bounds;
        }
        for(i = start; i <= end; i++)
            at(i) += f;
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::rotate_segment(int segment_index, ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        int            i,
                       out_of_bounds,
                       start,
                       end;
        ez_3D_short    center;

        find_segment(segment_index, start, end);
        center = find_segment_center(segment_index);

        for(i = start; i <= end; i++)
        {
            out_of_bounds = ez_ilda_bounds_check(rotate_point_around_point(at(i), center, a), EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        for(i = start; i <= end; i++)
            at(i) = ez_ilda_point( rotate_point_around_point(at(i), center, a),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::rotate_segment_around_origin(int segment_index, ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        int            i,
                       out_of_bounds,
                       start,
                       end;
        find_segment(segment_index, start, end);
        for(i = start; i <= end; i++)
        {
            out_of_bounds = ez_ilda_bounds_check(rotate_point(at(i), a), EZ_ILDA_SPHERE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        for(i = start; i <= end; i++)
            at(i) = ez_ilda_point( rotate_point(at(i), a),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::scale_segment(int segment_index, ez_3D_float m)
{
    if(number_of_points() > 1)
    {
        int            i,
                       out_of_bounds,
                       start,
                       end;
        ez_3D_short    center;

        find_segment(segment_index, start, end);
        center = find_segment_center(segment_index);

        for(i = start; i <= end; i++)
        {
            out_of_bounds = ez_ilda_bounds_check(scale_point_around_point(at(i), center, m), EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }
        for(i = start; i <= end; i++)
            at(i) = ez_ilda_point( scale_point_around_point(at(i), center, m),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c);
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_segment::scale_segment_around_origin(int segment_index, ez_3D_float m)
{
    if(number_of_points() > 1)
    {
        int            i,
                       out_of_bounds,
                       start,
                       end;

        find_segment(segment_index, start, end);
        for(i = start; i <= end; i++)
        {
            out_of_bounds = ez_ilda_bounds_check(m * at(i), EZ_ILDA_CUBE);
            if(out_of_bounds)
                return out_of_bounds;
        }

        for(i = start; i <= end; i++)
            at(i) = ez_ilda_point( m * at(i),
                                   (ez_color)at(i),
                                   at(i).k,
                                   at(i).c
                                 );
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
void ez_ilda_segment::to_ofstream_wave( ofstream&   out,
                                        const bool& end_of_frame,
                                        const bool& unique_frame
                                      ) const
{
    if(number_of_points() > 1)
    {
        int i;
        for(i = 0; i < number_of_points() - 1; i++)
            at(i).to_ofstream_wave(out, false, false);
        at(i).to_ofstream_wave(out, end_of_frame, unique_frame);
    }
    return;
}

//############################################################################






//############################################################################
ez_ilda_segment blank_segment()
{
    ez_ilda_segment segment(EZ_ILDA_DEFAULT);
    segment.push_back(ez_ilda_point(    0,     0, 0, 255, 255, 255, 64, 55));
    segment.push_back(ez_ilda_point(10000, 10000, 0, 255, 255, 255,  0, 55));
    segment.push_back(ez_ilda_point(    0,     0, 0, 255, 255, 255, 64, 55));
    return segment;
}

//############################################################################
ez_ilda_segment NULL_segment()
{
    ez_ilda_segment segment(EZ_EZFB_HUES);
    segment.push_back(ez_ilda_point(-19976, -8000, 0, 255, 255, 255, 64, 255));
    segment.push_back(ez_ilda_point(-19976, 10000, 0, 255, 255, 255,  0, 255));
    segment.push_back(ez_ilda_point( -9976, -8000, 0, 255, 255, 255,  0, 255));
    segment.push_back(ez_ilda_point( -9976, 10000, 0, 255, 255, 255,  0, 255));
    segment.push_back(ez_ilda_point( -7976, 10000, 0, 255, 255, 255, 64, 255));
    segment.push_back(ez_ilda_point( -7976, -6000, 0, 255, 255, 255,  0, 255));
    segment.push_back(ez_ilda_point( -5976, -8000, 0, 255, 255, 255,  0, 255));
    segment.push_back(ez_ilda_point( -1976, -8000, 0, 255, 255, 255,  0, 255));
    segment.push_back(ez_ilda_point(    24, -6000, 0, 255, 255, 255,  0, 255));
    segment.push_back(ez_ilda_point(    24, 10000, 0, 255, 255, 255,  0, 255));
    segment.push_back(ez_ilda_point(  2024, 10000, 0, 255, 255, 255, 64, 255));
    segment.push_back(ez_ilda_point(  2024, -8000, 0, 255, 255, 255,  0, 255));
    segment.push_back(ez_ilda_point( 10024, -8000, 0, 255, 255, 255,  0, 255));
    segment.push_back(ez_ilda_point( 12024, 10000, 0, 255, 255, 255, 64, 255));
    segment.push_back(ez_ilda_point( 12024, -8000, 0, 255, 255, 255,  0, 255));
    segment.push_back(ez_ilda_point( 20024, -8000, 0, 255, 255, 255,  0, 255));
    segment.push_back(ez_ilda_point(-19976, -8000, 0, 255, 255, 255, 64, 255));
    return segment;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
